@@ -155,7 +155,6 @@ sub new {
 		$self->_system_state = Cupt::System::State->new($self->_config, $self);
 	}
 
-	my @index_files;
 	foreach my $ref_index_entry (@$ref_index_entries) {
 		my $index_file_to_parse = $self->get_path_of_index_list($ref_index_entry);
 		my $source_type = $ref_index_entry->{'type'};
@@ -184,7 +183,6 @@ sub new {
 
 				$self->_process_index_file($index_file_to_parse, $chosen_translation_file,
 						$source_type, $ref_release_info);
-				push @index_files, $index_file_to_parse;
 			};
 			if (mycatch()) {
 				mywarn("skipped index file '%s'", $index_file_to_parse);
@@ -404,13 +402,20 @@ sub get_pin ($$) {
 
 	# discourage downgrading 
 	# downgradings will usually have pin <= 0
-	my $package_name = $version->package_name;
 	if (defined $self->_system_state) { # for example, for source versions will return false...
-		my $installed_version_string = $self->_system_state->get_installed_version_string($package_name);
-		if (defined $installed_version_string
-			&& Cupt::Core::compare_version_strings($installed_version_string, $version->version_string) > 0)
-		{
-			$result -= 2000;
+		my $installed_info = $self->_system_state->get_installed_info($version->package_name);
+		if (defined $installed_info) {
+			my $installed_version_string = $installed_info->{'version_string'};
+			if (defined $installed_version_string) {
+				if (Cupt::Core::compare_version_strings($installed_version_string, $version->version_string) > 0)
+				{
+					$result -= 2000;
+				}
+
+				if ($installed_info->{'want'} eq 'hold' && $version->is_installed()) {
+					$result += $self->_config->get_number('cupt::cache::obey-hold');
+				}
+			}
 		}
 	}
 
@@ -718,15 +723,24 @@ sub get_index_entries {
 
 sub __parse_source_list {
 	my ($file) = @_;
+
 	my @result;
-	open(HFILE, '<', $file) or mydie("unable to open file '%s': %s", $file, $!);
-	while (<HFILE>) {
+	open(my $fd, '<', $file) or mydie("unable to open file '%s': %s", $file, $!);
+	while (<$fd>) {
 		chomp;
 		# skip all empty lines and lines with comments
 		next if m/^\s*(?:#.*)?$/;
 
 		my %entry;
 		($entry{'type'}, $entry{'uri'}, $entry{'distribution'}, my @sections) = split ' ';
+
+		defined $entry{'type'} or
+				mydie("undefined source type at file '%s', line %u", $file, $.);
+		defined $entry{'uri'} or
+				mydie("undefined source URI at file '%s', line %u", $file, $.);
+		defined $entry{'distribution'} or
+				mydie("undefined source distribution at file '%s', line %u", $file, $.);
+
 		$entry{'uri'} =~ s{/$}{}; # strip last '/' if present
 
 		mydie("incorrect source type at file '%s', line %u", $file, $.)
@@ -750,7 +764,8 @@ sub __parse_source_list {
 			push @result, { %entry };
 		}
 	}
-	close(HFILE) or mydie("unable to close file '%s': %s", $file, $!);
+	close($fd) or mydie("unable to close file '%s': %s", $file, $!);
+
 	return @result;
 }
 
@@ -759,7 +774,7 @@ sub _parse_preference_file {
 
 	# we are parsing triads like:
 
-	# Package: perl
+	# Package: perl perl-modules
 	# Pin: o=debian,a=unstable
 	# Pin-Priority: 800
 
@@ -785,6 +800,7 @@ sub _parse_preference_file {
 			my $name_type = ($1 eq 'Package' ? 'package_name' : 'source_package_name');
 			my $name_value = $2;
 			glob_to_regex($name_value);
+			$name_value =~ s'\\ '|'g; # converting space separators into regex's OR
 
 			$pin_result{$name_type} = $name_value;
 		};
@@ -867,42 +883,36 @@ sub _parse_extended_states {
 	# Package: perl
 	# Auto-Installed: 1
 
+	# but, rarely another fields may be present, we need to ignore them
+
 	eval {
 		my $package_name;
 		my $value;
 
-		open(STATES, '<', $file) or mydie("unable to open file '%s': %s", $file, $!);
-		while (<STATES>) {
+		local $/ = "\n\n";
+
+		open(my $fd, '<', $file) or mydie("unable to open file '%s': %s", $file, $!);
+		while (<$fd>) {
+			m/^\s*(?:#.*|)$/ and next; # skip empty/comment lines
+
+			m'^Package: (.*?)$.*?^Auto-Installed: (.*?)$'sm or
+					mydie("bad chunk '%s' at file '%s'", $_, $file);
 			chomp;
 
-			# skipping newlines
-			next if $_ eq '';
+			$package_name = $1;
+			$value = $2;
 
-			do { # processing first line
-				m/^Package: (.*)/ or
-						mydie("bad package line at file '%s', line %u", $file, $.);
-
-				$package_name = $1;
-			};
-
-			do { # processing second line
-				my $value_line = <STATES>;
-				defined($value_line) or
-						mydie("no value line at file '%s' line %u", $file, $.);
-
-				$value_line =~ m/^Auto-Installed: (0|1)/ or
-						mydie("bad value line at file '%s' line %u", $file, $.);
-
-				$value = $1;
-			};
+			if ($value ne '0' and $value ne '1') {
+				mydie("bad value '%s' (should be 0 or 1) in chunk '%s' at file '%s'",
+						$value, $_, $file);
+			}
 
 			if ($value) {
 				# adding to storage
 				$self->[_extended_info_offset()]->{'automatically_installed'}->{$package_name} = $value;
 			}
 		}
-
-		close(STATES) or mydie("unable to close file '%s': %s", $file, $!);
+		close($fd) or mydie("unable to close file '%s': %s", $file, $!);
 	};
 	if (mycatch()) {
 		myerr('error while parsing extended states');
@@ -954,6 +964,9 @@ sub _process_index_file {
 			my $offset = tell($fh) - length($_);
 			my ($package_name) = m/^Package: (.*?)$/m;
 
+			defined $package_name or
+					mydie("unable to find package name");
+
 			# we skips 'Package: <...>' line additionally
 			$offset += length('Package: ') + length($package_name) + 1;
 
@@ -1288,6 +1301,13 @@ sub _parse_preferences {
 	my $parts_dir = $self->_config->get_string('dir::etc::preferencesparts');
 	my @preference_files = glob("$root_prefix$etc_dir/$parts_dir/*");
 
+	@preference_files = grep {
+		(my $name = $_) =~ s{.*/}{}; # cutting directory name
+		$name =~ m/^[A-Za-z0-9_.-]+$/
+			and
+		($name !~ m/\./ or $name =~ m/\.pref$/) # an extension, if exist, should be 'pref'
+	} @preference_files;
+
 	my $main_file = $self->_config->get_string('dir::etc::preferences');
 	my $main_file_path = "$root_prefix$etc_dir/$main_file";
 	push @preference_files, $main_file_path if -r $main_file_path;
@@ -1369,57 +1389,64 @@ sub verify_signature ($$) {
 
 	my $signature_file = "$file.gpg";
 	mydebug("signature file is '%s'", $signature_file) if $debug;
+
 	-r $signature_file or
 			do {
 				mydebug("unable to read signature file '%s'", $signature_file) if $debug;
 				return 0;
 			};
 
-	-e $keyring_file or
-			do {
-				mywarn("keyring file '%s' doesn't exist", $keyring_file);
-				return 0;
-			};
-	-r $keyring_file or
-			do {
-				mywarn("no read rights on keyring file '%s', please do 'chmod +r %s' with root rights",
-						$keyring_file, $keyring_file);
-				return 0;
-			};
-
-	open(GPG_VERIFY, 'gpgv --status-fd 1 ' .
-			"--keyring $keyring_file $signature_file $file 2>/dev/null |") or
-			mydie('unable to open gpg pipe: %s', $!);
-	my $sub_gpg_readline = sub {
-		my $result;
-		do {
-			$result = readline(GPG_VERIFY);
-			if (defined $result) {
-				chomp $result;
-				mydebug("fetched '%s' from gpg pipe", $result) if $debug;
-			}
-		} while (defined $result and (($result =~ m/^\[GNUPG:\] SIG_ID/) or ($result !~ m/^\[GNUPG:\]/)));
-
-		if (!defined $result) {
-			return undef;
-		} else {
-			$result =~ s/^\[GNUPG:\] //;
-			return $result;
-		}
-	};
 	my $verify_result;
+	eval {
+		-e $keyring_file or
+				mydie("keyring file '%s' doesn't exist", $keyring_file);
+		-r $keyring_file or
+				mydie("no read rights on keyring file '%s', please do 'chmod +r %s' with root rights",
+							$keyring_file, $keyring_file);
+
+		open(GPG_VERIFY, 'gpgv --status-fd 1 ' .
+				"--keyring $keyring_file $signature_file $file 2>/dev/null |") or
+				mydie('unable to open gpg pipe: %s', $!);
+		my $sub_gpg_readline = sub {
+			my $result;
+			do {
+				$result = readline(GPG_VERIFY);
+				if (defined $result) {
+					chomp $result;
+					mydebug("fetched '%s' from gpg pipe", $result) if $debug;
+				}
+			} while (defined $result and (($result =~ m/^\[GNUPG:\] SIG_ID/) or ($result !~ m/^\[GNUPG:\]/)));
+
+			if (!defined $result) {
+				return undef;
+			} else {
+				$result =~ s/^\[GNUPG:\] //;
+				return $result;
+			}
+		};
+
+		my $status_string = $sub_gpg_readline->();
+		if (not defined $status_string) {
+			# no info from gpg at all
+			mydie("gpg: '%s': no info received", $file);
+		}
 
-	my $status_string = $sub_gpg_readline->();
-	if (defined $status_string) {
 		# first line ought to be validness indicator
 		my ($message_type, $message) = ($status_string =~ m/(\w+) (.*)/);
+		if (not defined $message_type or not defined $message) {
+			mydie("gpg: '%s': invalid status string '%s'", $file, $status_string);
+		}
+
 		given ($message_type) {
 			when ('GOODSIG') {
 				my $further_info = $sub_gpg_readline->();
 				defined $further_info or
-						mydie("gpg: '%s': unfinished status");
+						mydie("gpg: '%s': error: unfinished status", $file);
 
 				my ($check_result_type, $check_message) = ($further_info =~ m/(\w+) (.*)/);
+				if (not defined $check_result_type or not defined $check_message) {
+					mydie("gpg: '%s': invalid further info string '%s'", $file, $further_info);
+				}
 				given ($check_result_type) {
 					when ('VALIDSIG') {
 						# no comments :)
@@ -1438,7 +1465,7 @@ sub verify_signature ($$) {
 						mywarn("gpg: '%s': revoked key: %s", $file, $check_message);
 					}
 					default {
-						mydie("gpg: '%s': unknown error: %s %s", $file, $check_result_type, $check_message);
+						mywarn("gpg: '%s': unknown error: %s %s", $file, $check_result_type, $check_message);
 					}
 				}
 			}
@@ -1448,27 +1475,66 @@ sub verify_signature ($$) {
 			}
 			when ('ERRSIG') {
 				# gpg was not able to verify signature
-				mywarn("gpg: '%s': could not verify signature: %s", $file, $message);
 				$verify_result = 0;
+
+				# maybe, public key was not found?
+				my $public_key_was_not_found = 0;
+				my $detail_string = $sub_gpg_readline->();
+				if (defined $detail_string) {
+					my ($detail_type, $detail_message) = ($detail_string =~ m/(\w+) (.*)/);
+					if (not defined $detail_type or not defined $detail_message) {
+						mydie("gpg: '%s': invalid detailed info string '%s'", $file, $detail_string);
+					}
+					if ($detail_type eq 'NO_PUBKEY') {
+						$public_key_was_not_found = 1;
+
+						# the message looks like
+						#
+						# NO_PUBKEY D4F5CE00FA0E9B9D
+						#
+						# the best guess for getting the public key is second 8 hex digits within this string
+						my $public_key_number_detected = 0;
+						if (length($detail_message) == 16) {
+							my $public_key_number = substr($detail_message, 8, 8);
+							if ($public_key_number =~ m/^[0-9A-Z]{8}$/) {
+								$public_key_number_detected = 1;
+								mywarn("gpg: '%s': public key '%s' not found", $file, $public_key_number);
+							}
+						}
+
+						if (not $public_key_number_detected) {
+							mywarn("gpg: '%s': public key not found", $file);
+						}
+					}
+				}
+
+				if (not $public_key_was_not_found) {
+					mywarn("gpg: '%s': could not verify signature: %s", $file, $message);
+				}
 			}
 			when ('NODATA') {
 				# no signature
 				mywarn("gpg: '%s': empty signature", $file);
 				$verify_result = 0;
 			}
+			when ('KEYEXPIRED') {
+				$verify_result = 0;
+				mywarn("gpg: '%s': expired key: %s", $file, $message);
+			}
 			default {
-				mydie("gpg: '%s': unknown message received: %s %s", $file, $message_type, $message);
+				mywarn("gpg: '%s': unknown message received: %s %s", $file, $message_type, $message);
+				$verify_result = 0;
 			}
 		}
-	} else {
-		# no info from gpg at all
+
+		close(GPG_VERIFY) or $! == 0 or
+				mydie('unable to close gpg pipe: %s', $!);
+	};
+	if (mycatch()) {
 		mywarn("error while verifying signature for file '%s'", $file);
 		$verify_result = 0;
 	}
 
-	close(GPG_VERIFY) or $! == 0 or
-			mydie('unable to close gpg pipe: %s', $!);
-
 	mydebug('the verify result is %u', $verify_result) if $debug;
 	return $verify_result;
 }
