@@ -62,7 +62,7 @@ my $grammar = q{
 	      '#clear' name ';'
 
 	name:
-		  /([\w\/-]+::)*([\w\/-]+)/
+		  /([\w\/.-]+::)*([\w\/.-]+)/
 
 	value:
 		  /".*"/
