@@ -34,7 +34,7 @@ use IO::Select;
 use IO::Pipe;
 use IO::Socket::UNIX;
 use POSIX;
-use Time::HiRes qw(setitimer ITIMER_REAL);
+use Time::HiRes qw(setitimer ITIMER_VIRTUAL);
 use URI;
 
 use Cupt::LValueFields qw(_config _progress _worker_pid _server_socket
@@ -160,10 +160,18 @@ sub _worker ($) {
 	my $exit_flag = 0;
 
 	# setting progress ping timer
-	local $SIG{ALRM} = sub { __my_write_socket($worker_writer, 'progress', '', 'ping') };
-	setitimer(ITIMER_REAL, 0.25, 0.25);
+	#
+	# unfortunately, Perl has no signalfd() wrapper so we can use it in IO::Select dispatcher
+	# so, to avoid potential clash with simulateneously writing to $worker_writer in a signal
+	# and in the processing thread, create another pair of sockets
+	pipe(my $worker_ping_reader, my $worker_ping_writer);
+	$worker_ping_writer->autoflush(1);
+	local $SIG{VTALRM} = sub {
+		__my_write_socket($worker_ping_writer, 'progress', '', 'ping')
+	};
+	setitimer(ITIMER_VIRTUAL, 0.25, 0.25);
 
-	my @persistent_sockets = ($worker_reader, $self->_parent_pipe, $self->_server_socket);
+	my @persistent_sockets = ($worker_reader, $worker_ping_reader, $self->_parent_pipe, $self->_server_socket);
 	my @runtime_sockets;
 
 	# while caller may set exit flag, we should continue processing as long as
@@ -206,7 +214,7 @@ sub _worker ($) {
 					# new query appeared
 					($uri, $filename) = @params;
 					$waiter_socket = $socket;
-					mydebug("download request: '$uri'") if $debug;
+					mydebug("download request: '%s'", $uri) if $debug;
 					$proceed_next_download = 1;
 				}
 				when ('set-download-size') {
@@ -219,7 +227,7 @@ sub _worker ($) {
 							myinternaldie("bad argument count for 'done' message");
 
 					($uri, my $result) = @params;
-					mydebug("preliminary download result: '$uri': $result") if $debug;
+					mydebug("preliminary download result: '%s': $result", $uri) if $debug;
 					my $is_duplicated_download = 0;
 					__my_write_socket($active_downloads{$uri}->{waiter_socket}, $uri, $result, $is_duplicated_download);
 
@@ -235,7 +243,7 @@ sub _worker ($) {
 							myinternaldie("bad argument count for 'done-ack' message");
 
 					($uri, my $result) = @params;
-					mydebug("final download result: '$uri': $result") if $debug;
+					mydebug("final download result: '%s': $result", $uri) if $debug;
 
 					# removing the query from active download list and put it to
 					# the list of ended ones
@@ -250,7 +258,7 @@ sub _worker ($) {
 						foreach my $ref_pending_download (@pending_downloads) {
 							(my $pending_uri, $waiter_socket) = @$ref_pending_download;
 							if (exists $done_downloads{$pending_uri}) {
-								mydebug("final download result for duplicated request: '$pending_uri': $result") if $debug;
+								mydebug("final download result for duplicated request: '%s': $result", $pending_uri) if $debug;
 								__my_write_socket($waiter_socket, $pending_uri, $result, $is_duplicated_download);
 								close($waiter_socket) or
 										mydie("unable to close waiter socket for uri '%s'", $pending_uri);
@@ -297,7 +305,7 @@ sub _worker ($) {
 					if (scalar @download_queue) {
 						# put next of waiting queries
 						($uri, $filename, $waiter_socket) = @{shift @download_queue};
-						mydebug("enqueue '$uri' from hold") if $debug;
+						mydebug("enqueue '%s' from hold", $uri) if $debug;
 						$proceed_next_download = 1;
 					}
 				}
@@ -311,12 +319,12 @@ sub _worker ($) {
 			}
 
 			$proceed_next_download or next;
-			mydebug("processing download '$uri'") if $debug;
+			mydebug("processing download '%s'", $uri) if $debug;
 
 			# check if this download was already done
 			if (exists $done_downloads{$uri}) {
 				my $result = $done_downloads{$uri};
-				mydebug("final result for duplicated download '$uri': $result") if $debug;
+				mydebug("final result for duplicated download '%s': $result", $uri) if $debug;
 				# just immediately end it
 				my $is_duplicated_download = 1;
 				__my_write_socket($waiter_socket, $uri, $result, $is_duplicated_download);
@@ -325,19 +333,19 @@ sub _worker ($) {
 				__my_write_socket($worker_writer, 'pop-download');
 				next;
 			} elsif (exists $active_downloads{$uri}) {
-				mydebug("pushed '$uri' to pending queue") if $debug;
+				mydebug("pushed '%s' to pending queue", $uri) if $debug;
 				push @pending_downloads, [ $uri, $waiter_socket ];
 				next;
 			} elsif (scalar keys %active_downloads >= $max_simultaneous_downloads_allowed) {
 				# put the query on hold
-				mydebug("put '$uri' on hold") if $debug;
+				mydebug("put '%s' on hold", $uri) if $debug;
 				push @download_queue, [ $uri, $filename, $waiter_socket ];
 				next;
 			}
 
 			# there is a space for new download, start it
 
-			mydebug("starting download '$uri'") if $debug;
+			mydebug("starting download '%s'", $uri) if $debug;
 			$target_filenames{$uri} = $filename;
 			# filling the active downloads hash
 			$active_downloads{$uri}->{waiter_socket} = $waiter_socket;
@@ -377,13 +385,15 @@ sub _worker ($) {
 		}
 	}
 	# disabling timer
-	local $SIG{ALRM} = sub {};
-	setitimer(ITIMER_REAL, 0, 0);
+	local $SIG{VTALRM} = sub {};
+	setitimer(ITIMER_VIRTUAL, 0, 0);
 	# finishing progress
 	$self->_progress->finish();
 
 	close($worker_reader) or mydie("unable to close worker's own reader socket: %s", $!);
 	close($worker_writer) or mydie("unable to close worker's own writer socket: %s", $!);
+	close($worker_ping_reader) or mydie("unable to close worker's ping reader socket: %s", $!);
+	close($worker_ping_writer) or mydie("unable to close worker's ping writer socket: %s", $!);
 	mydebug('download worker process finished') if $debug;
 	POSIX::_exit(0);
 	return;
