@@ -49,7 +49,7 @@ use Cupt::System::Worker::Lock;
 
 my $_download_partial_suffix = '/partial';
 
-use Cupt::LValueFields qw(_config _cache _system_state _desired_state _lock);
+use Cupt::LValueFields qw(_config _cache _system_state _desired_state _lock _umask);
 
 =head1 METHODS
 
@@ -72,6 +72,8 @@ sub new {
 	$self->_cache = shift;
 	$self->_system_state = $self->_cache->get_system_state();
 	$self->_desired_state = undef;
+
+	$self->_umask = umask(0022);
 	$self->_lock = Cupt::System::Worker::Lock->new($self->_config,
 			$self->_config->get_string('dir') . $self->_config->get_string('cupt::directory::state') . '/lock');
 	$self->_lock->obtain();
@@ -85,6 +87,7 @@ sub DESTROY {
 	my ($self) = @_;
 
 	$self->_lock->release();
+	umask($self->_umask);
 
 	return;
 }
@@ -502,6 +505,10 @@ sub _fill_actions ($$\@) {
 		'purge' => [ 'remove' ],
 	);
 
+	# packages with indirect upgrades have explicit 'remove' action, so
+	# their 'remove' subactions are not fake
+	my @fake_exception_package_names = $self->_config->get_list('cupt::worker::allow-indirect-upgrade');
+
 	# convert all actions into inner ones
 	foreach my $user_action (keys %$ref_actions_preview) {
 		my $ref_actions_to_be_performed = $user_action_to_inner_actions{$user_action};
@@ -536,9 +543,12 @@ sub _fill_actions ($$\@) {
 					'action_name' => defined $version ? $inner_action : 'purge-config-files',
 				};
 				if ($inner_action eq 'remove' and scalar @$ref_actions_to_be_performed > 1) {
-					# this is fake 'remove' action which will be merged
-					# with 'unpack' unconditionally later
-					$ref_action->{'fake'} = 1;
+					my $package_name = $ref_package_entry->{'package_name'};
+					if (none { $_ eq $package_name } @fake_exception_package_names) {
+						# this is fake 'remove' action which will be merged
+						# with 'unpack' unconditionally later
+						$ref_action->{'fake'} = 1;
+					}
 				}
 
 				$graph->add_vertex($ref_action);
@@ -573,6 +583,7 @@ sub _fill_action_dependencies ($$$$) {
 			);
 			# search for the appropriate action in action list
 			my $ref_search_index = $graph->get_graph_attribute('vertexes_by_package_name');
+			CURRENT_ACTION:
 			foreach my $ref_current_action (@{$ref_search_index->{$other_version->package_name}}) {
 				next if exists $ref_inner_action->{'fake'} and exists $ref_current_action->{'fake'};
 				if (__is_inner_actions_equal(\%candidate_action, $ref_current_action)) {
@@ -596,7 +607,9 @@ sub _fill_action_dependencies ($$$$) {
 					if ($self->_config->get_bool('debug::worker')) {
 						my $slave_string = __stringify_inner_action($ref_slave_action);
 						my $master_string = __stringify_inner_action($ref_master_action);
-						mydebug("new action dependency: '$slave_string' -> '$master_string'");
+						my $reason = $dependency_name . ': ' .
+								stringify_relation_expression($relation_expression);
+						mydebug("new action dependency: '$slave_string' -> '$master_string', reason: '$reason'");
 					}
 
 					next SATISFYING_VERSIONS;
@@ -607,79 +620,8 @@ sub _fill_action_dependencies ($$$$) {
 	return;
 }
 
-sub _build_actions_graph ($$) {
-	my ($self, $ref_actions_preview) = @_;
-
-	if (!defined $self->_desired_state) {
-		myinternaldie('worker desired state is not given');
-	}
-
-	# action = {
-	# 	'package_name' => package
-	# 	'version_string' => version_string,
-	# 	'action_name' => ('unpack' | 'configure' | 'remove')
-	# }
-	my $graph = Cupt::Graph->new();
-
-	$self->_fill_actions($ref_actions_preview, $graph);
-
-	# maybe, we have nothing to do?
-	return undef if scalar $graph->vertices() == 0;
-
-	# here we also adding adding fake-antiupgrades for all packages in the system
-	# which stay unmodified for the sake of getting inter-dependencies between
-	# the optional dependecies like
-	#
-	# 'abc' depends 'x | y', abc stays unmodified, x goes away, y is going to be installed
-	#
-	# here, action 'remove x' are dependent on 'install y' one and it gets
-	# introduced by
-	#
-	# 'install y' -> 'install abc' <-> 'remove abc' -> 'remove x'
-	#                {----------------------------}
-	#                            <merge>
-	# which goes into
-	#
-	# 'install y' -> 'remove x'
-	#
-	# moreover, we split the installed version into virtual version by one
-	# relation expression, so different relation expressions of the same real
-	# version don't interact with each other, otherwise we'd get a full cyclic
-	# mess
-	my @virtual_edges_to_be_eliminated;
-	do {
-		my @blacklisted_package_names;
-		# the black list
-		foreach my $ref_inner_action ($graph->vertices()) {
-			push @blacklisted_package_names, $ref_inner_action->{'version'}->package_name;
-		}
-		foreach my $version (@{$self->_cache->get_system_state()->export_installed_versions()}) {
-			my $package_name = $version->package_name;
-			next if any { $package_name eq $_ } @blacklisted_package_names;
-
-			foreach my $dependency_name (qw(pre_depends depends)) {
-				foreach my $relation_expression (@{$version->$dependency_name}) {
-					my $virtual_version = (bless [] => 'Cupt::Cache::BinaryVersion');
-					$virtual_version->package_name = "$package_name [" .
-							stringify_relation_expression($relation_expression) . ']';
-					$virtual_version->version_string = $version->version_string;
-					$virtual_version->pre_depends = [];
-					$virtual_version->depends = [];
-					$virtual_version->$dependency_name = [ $relation_expression ];
-					$virtual_version->conflicts = [];
-					$virtual_version->breaks = [];
-
-					my $from_vertex = { 'version' => $virtual_version, 'action_name' => 'configure', 'fake' => 1 };
-					my $to_vertex = { 'version' => $virtual_version, 'action_name' => 'remove', 'fake' => 1 };
-					# we don't add edge here, but add the vertexes to gain dependencies and
-					# save the vertexes order
-					$graph->add_vertex($from_vertex);
-					$graph->add_vertex($to_vertex);
-					push @virtual_edges_to_be_eliminated, [ $from_vertex, $to_vertex ];
-				}
-			}
-		}
-	};
+sub _fill_graph_dependencies {
+	my ($self, $graph) = @_;
 
 	my %vertexes_by_package_name;
 	do { # building the action index for fast action search
@@ -761,6 +703,114 @@ sub _build_actions_graph ($$) {
 		}
 	}
 	$graph->delete_graph_attributes();
+}
+
+sub __check_graph_pre_depends {
+	my ($graph) = @_;
+
+	# re-compute transitive matrix, it might become invalid after the merges
+	my $graph_transitive_closure = Cupt::Graph::TransitiveClosure->new($graph);
+
+	foreach my $edge ($graph->edges()) {
+		my $ref_dependency_expressions = $graph->get_edge_attribute(@$edge, 'dependency_expressions');
+		next if not defined $ref_dependency_expressions;
+
+		my @pre_dependency_relation_expressions = map { $_->{'relation_expression'} }
+				(grep { $_->{'name'} eq 'pre_depends' } @$ref_dependency_expressions);
+		next unless scalar @pre_dependency_relation_expressions;
+
+		my ($from_vertex, $to_vertex) = @$edge;
+		if ($graph_transitive_closure->is_reachable($to_vertex, $from_vertex)) {
+			# bah! the pre-dependency cannot be overridden, it's a fatal error
+			# which is not worker's fail (at least, it shouldn't be)
+
+			my @path = $graph_transitive_closure->path_vertices($to_vertex, $from_vertex);
+			my @package_names_in_path = uniq map { $_->{'version'}->package_name } @path;
+
+			mywarn("the pre-dependency(ies) '%s' will be broken during the actions, the packages involved: %s",
+					stringify_relation_expressions(\@pre_dependency_relation_expressions),
+					join(', ', map { qq/'$_'/ } @package_names_in_path));
+		}
+	}
+
+	return undef;
+}
+
+sub _build_actions_graph ($$) {
+	my ($self, $ref_actions_preview) = @_;
+
+	if (!defined $self->_desired_state) {
+		myinternaldie('worker desired state is not given');
+	}
+
+	# action = {
+	# 	'package_name' => package
+	# 	'version_string' => version_string,
+	# 	'action_name' => ('unpack' | 'configure' | 'remove')
+	# }
+	my $graph = Cupt::Graph->new();
+
+	$self->_fill_actions($ref_actions_preview, $graph);
+
+	# maybe, we have nothing to do?
+	return undef if scalar $graph->vertices() == 0;
+
+	# here we also adding adding fake-antiupgrades for all packages in the system
+	# which stay unmodified for the sake of getting inter-dependencies between
+	# the optional dependecies like
+	#
+	# 'abc' depends 'x | y', abc stays unmodified, x goes away, y is going to be installed
+	#
+	# here, action 'remove x' are dependent on 'install y' one and it gets
+	# introduced by
+	#
+	# 'install y' -> 'install abc' <-> 'remove abc' -> 'remove x'
+	#                {----------------------------}
+	#                            <merge>
+	# which goes into
+	#
+	# 'install y' -> 'remove x'
+	#
+	# moreover, we split the installed version into virtual version by one
+	# relation expression, so different relation expressions of the same real
+	# version don't interact with each other, otherwise we'd get a full cyclic
+	# mess
+	my @virtual_edges_to_be_eliminated;
+	do {
+		my @blacklisted_package_names;
+		# the black list
+		foreach my $ref_inner_action ($graph->vertices()) {
+			push @blacklisted_package_names, $ref_inner_action->{'version'}->package_name;
+		}
+		foreach my $version (@{$self->_cache->get_system_state()->export_installed_versions()}) {
+			my $package_name = $version->package_name;
+			next if any { $package_name eq $_ } @blacklisted_package_names;
+
+			foreach my $dependency_name (qw(pre_depends depends)) {
+				foreach my $relation_expression (@{$version->$dependency_name}) {
+					my $virtual_version = (bless [] => 'Cupt::Cache::BinaryVersion');
+					$virtual_version->package_name = "$package_name [" .
+							stringify_relation_expression($relation_expression) . ']';
+					$virtual_version->version_string = $version->version_string;
+					$virtual_version->pre_depends = [];
+					$virtual_version->depends = [];
+					$virtual_version->$dependency_name = [ $relation_expression ];
+					$virtual_version->conflicts = [];
+					$virtual_version->breaks = [];
+
+					my $from_vertex = { 'version' => $virtual_version, 'action_name' => 'configure', 'fake' => 1 };
+					my $to_vertex = { 'version' => $virtual_version, 'action_name' => 'remove', 'fake' => 1 };
+					# we don't add edge here, but add the vertexes to gain dependencies and
+					# save the vertexes order
+					$graph->add_vertex($from_vertex);
+					$graph->add_vertex($to_vertex);
+					push @virtual_edges_to_be_eliminated, [ $from_vertex, $to_vertex ];
+				}
+			}
+		}
+	};
+
+	$self->_fill_graph_dependencies($graph);
 
 	do { # unit all downgrades/upgrades
 		# list of packages affected
@@ -855,8 +905,6 @@ sub _build_actions_graph ($$) {
 			$graph->delete_vertex($from);
 		}
 
-		my $graph_transitive_closure = Cupt::Graph::TransitiveClosure->new($graph);
-
 		if ($self->_config->get_bool('debug::worker')) {
 			foreach my $edge ($graph->edges()) {
 				my $slave_string = __stringify_inner_action($edge->[0]);
@@ -865,32 +913,7 @@ sub _build_actions_graph ($$) {
 			}
 		}
 
-		do { # check pre-depends
-			# re-compute transitive matrix, it might become invalid after the merges
-			$graph_transitive_closure = Cupt::Graph::TransitiveClosure->new($graph);
-
-			foreach my $edge ($graph->edges()) {
-				my $ref_dependency_expressions = $graph->get_edge_attribute(@$edge, 'dependency_expressions');
-				next if not defined $ref_dependency_expressions;
-
-				my @pre_dependency_relation_expressions = map { $_->{'relation_expression'} }
-						(grep { $_->{'name'} eq 'pre_depends' } @$ref_dependency_expressions);
-				next unless scalar @pre_dependency_relation_expressions;
-
-				my ($from_vertex, $to_vertex) = @$edge;
-				if ($graph_transitive_closure->is_reachable($to_vertex, $from_vertex)) {
-					# bah! the pre-dependency cannot be overridden, it's a fatal error
-					# which is not worker's fail (at least, it shouldn't be)
-
-					my @path = $graph_transitive_closure->path_vertices($to_vertex, $from_vertex);
-					my @package_names_in_path = uniq map { $_->{'version'}->package_name } @path;
-
-					mywarn("the pre-dependency(ies) '%s' will be broken during the actions, the packages involved: %s",
-							stringify_relation_expressions(\@pre_dependency_relation_expressions),
-							join(', ', map { qq/'$_'/ } @package_names_in_path));
-				}
-			}
-		};
+		__check_graph_pre_depends($graph);
 	};
 
 	return $graph;
@@ -939,8 +962,9 @@ sub __split_heterogeneous_actions (@) {
 						any { $_->{'action_name'} ne $first_action_name } @$ref_action_subgroup)
 					{
 						# ooh, mixed circular dependency? no-go
-						mydie("unable to schedule circular actions '%s'",
+						myerr("unable to schedule circular actions '%s'",
 								join(', ', map { __stringify_inner_action($_) } @$ref_action_subgroup));
+						mydie("you may try to work around by adding one or more affected packages to the 'cupt::worker::allow-indirect-upgrade' configuration variable");
 					}
 				}
 
@@ -1206,13 +1230,10 @@ sub _generate_stdin_for_preinstall_hooks_version2 ($$) {
 			} elsif ($action_name eq 'remove') {
 				# and for this...
 				$filename = '**REMOVE**';
-			} else { # unpack or install
+			} else { # unpack
 				$filename = $self->_get_archives_directory() . '/' . __get_archive_basename($action_version);
 			}
-			$result .= "$package_name $old_version_string < $new_version_string $filename\n";
-			if ($action_name eq 'install') {
-				$result .= "$package_name $old_version_string < $new_version_string **CONFIGURE**\n";
-			}
+			$result .= "$package_name $old_version_string $compare_version_strings_sign $new_version_string $filename\n";
 		}
 	}
 
